/**
  *  \file client/billexport.cc
  *  \brief Exporting a Bill
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <stdexcept>
#include "client/billexport.h"
#include "client/actions/gaexport.h"
#include "client/dialogs/exportparam.h"
#include "cpluslib/misc.h"
#include "int/context.h"
#include "int/export.h"
#include "int/if/ifutil.h"
#include "int/interror.h"
#include "int/propenum.h"
#include "util/answer.h"

namespace {
    /*
     *  The "Export" infrastructure is built around interpreter contexts.
     *  Thus, the simplest way to re-use it is to build in interpreter context that describes a bill.
     *  This would normally bring up a plethora of lifetime issues,
     *  i.e. what if the script stores a reference to the bill in a global variable?
     *  I therefore use some knowledge about the innards of Export,
     *  in particular, that it never calls out into the interpreter
     *  in a way that would allow cloning the BillContext.
     */
    enum {
        bpName,
        bpCount,
        bpT,
        bpD,
        bpM,
        bpCash
    };

    const TIntMapping bill_mapping[] = {
        { "CASH",     bpCash,     0,  thInt },
        { "COUNT",    bpCount,    0,  thInt },
        { "D",        bpD,        0,  thInt },
        { "M",        bpM,        0,  thInt },
        { "NAME",     bpName,     0,  thString },
        { "T",        bpT,        0,  thInt },
    };

    const char DEFAULT_FORMAT[] = "COUNT@5,NAME@40,T@6,D@6,M@6,CASH@6";

    class BillContext : public IntContext {
     public:
        BillContext(const WBill& bill)
            : bill(bill),
              current(0)
            { }
        virtual int32_t lookup(const IntNameQuery& name)
            { return lookupName(name, bill_mapping, countof(bill_mapping)); }
        virtual void set(int32_t /*index*/, IntValue* /*value*/)
            { throw IntError::notAssignable(); }
        virtual IntValue* get(int32_t index)
            {
                if (current < bill.size()) {
                    const WBillItem& item = *bill[current];
                    switch (bill_mapping[index].index) {
                     case bpName:  return makeStringValue(item.item_name);
                     case bpCount: return makeIntValue(item.display_mult);
                     case bpT:     return makeIntValue(item.item_cost.get(el_Tritanium));
                     case bpD:     return makeIntValue(item.item_cost.get(el_Duranium));;
                     case bpM:     return makeIntValue(item.item_cost.get(el_Molybdenum));
                     case bpCash:  return makeIntValue(item.item_cost.get(el_Money) + item.item_cost.get(el_Supplies));
                    }
                }
                return 0;
            }
        virtual bool next()
            {
                ++current;
                return current < bill.size();
            }
        virtual IntContext* clone()
            { throw std::runtime_error("Not clonable"); }
        virtual GObject* getObject()
            { return 0; }
        virtual void enumProperties(IntPropertyAcceptor& acceptor)
            { acceptor.enumMapping(bill_mapping, countof(bill_mapping)); }
        string_t toString(bool /*readable*/)
            { return "#<bill>"; }
        void store(IntTagNode& /*sv*/, Stream& /*aux*/)
            { throw IntError::notSerializable(); }
     private:
        const WBill& bill;
        size_t current;
    };
}

/** Export a bill.
    Displays the dialog and performs the export. */
void
doBillExport(const WBill& bill)
{
    // Do not export empty bills.
    if (bill.empty()) {
        return;
    }

    // Export field list
    IntExportFieldList fields;
    try {
        fields.addList(DEFAULT_FORMAT);
    }
    catch (...)
    { }

    // Export object type
    BillContext ctx(bill);

    // Dialog
    WExportParameterDialog dialog(fields, &ctx, ConstantAnswerProvider::sayYes);
    ui_root->centerWidget(dialog);
    ui_root->runModal(dialog);
}
