/**
  *  \file client/screens/taskscreen.cc
  *  \brief Auto Task Screen
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens/taskscreen.h"
#include "client/tiles.h"
#include "int/exec.h"
#include "ui/msgbox.h"
#include "ui/widgets/inputline.h"
#include "util/nls.h"
#include "util/strlist.h"
#include "ui/listbox.h"
#include "ui/util.h"
#include "ui/dialogs/filedlg.h"
#include "io/textfile.h"
#include "util/except.h"
#include "cpluslib/format.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/widgets/statictext.h"
#include "int/context.h"
#include "client/dialogs/consoledlg.h"
#include "ui/widgets/quit.h"

namespace {
    class WEditCommandWindow : public UIWindow {
     public:
        WEditCommandWindow(const char* ttl);

        void onOk();
        void onCancel();

        string_t getText()
            { return input->getText(); }
        void setText(const string_t& str)
            { input->setText(str); }
        bool handleEvent(const UIEvent& e, bool second_pass);

     private:
        void init();

        ObjectHolder h;
        UIInputLine* input;
        ptr_vector<IntContext> contextCache;
    };
}

WEditCommandWindow::WEditCommandWindow(const char* ttl)
    : UIWindow(ttl, UIWindow::BLUE_WINDOW, COLOR_GRAY, 10)
{
    init();
}

void
WEditCommandWindow::onOk()
{
    if (!IntAutoTaskEditor::isValidCommand(input->getText())) {
        messageBox(_("This is not a valid auto task command."), _("Error"));
    } else {
        stop(1);
    }
}

void
WEditCommandWindow::onCancel()
{
    stop(0);
}

bool
WEditCommandWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key && e.key.code == SDLK_TAB) {
        doComplete(*input, contextCache, true);
        return true;
    } else {
        return UIWindow::handleEvent(e, second_pass);
    }
}

void
WEditCommandWindow::init()
{
    int width = std::min(40, (ui_root->getExtent().w - 40) / font_ems[FONT_TITLE]);

    setLayout(UIVBoxLayout::instance5);

    add(h.add(new UIStaticText(_("Command:"), UIStaticText::StaticTitle)));

    input = &h.add(new UIInputLine(4000, width, 0, 0, 0));
    add(*input);

    UIGroup& g(h.add(new UIGroup(UIHBoxLayout::instance5)));
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addCall(this, &WEditCommandWindow::onOk));
    g.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addCall(this, &WEditCommandWindow::onCancel));
    add(g);
    add(h.add(new UIQuit(0)));

    pack();

    IntUserInterfaceBinding::get().enumContexts(contextCache);
}

static void
editCommand(const string_t& cmd, const char* ttl, WAutoTaskObjectSelection& selection, int del)
{
    WEditCommandWindow win(ttl);
    win.setText(cmd);
    ui_root->centerWidget(win);
    if (ui_root->runModal(win)) {
        selection.replace(selection.getCursor(), del, win.getText());
    }
}

/** Constructor.
    \param base Underlying object type */
WAutoTaskObjectSelection::WAutoTaskObjectSelection(GObjectSelection& selection)
    : GProxyObjectSelection(),
      GObjectSelectionObserver(selection),
      current_process(0),
      cursor(0)
{
    GProxyObjectSelection::setObjectSelection(&selection);
    sig_current_changed.add(this, &WAutoTaskObjectSelection::onCurrentChanged);
}

WAutoTaskObjectSelection::~WAutoTaskObjectSelection()
{
    done(); /* just in case */
}

/** Initialize. Call before using the auto task editor screen, but after
    finishing construction (virtual methods must be available). */
void
WAutoTaskObjectSelection::init()
{
    onCurrentChanged();
}

/** Initialize. Call after using the auto task editor screen, but before
    starting destruction (virtual methods must be available). */
void
WAutoTaskObjectSelection::done()
{
    loadTask(0);
}

/** Set cursor. */
void
WAutoTaskObjectSelection::setCursor(uint32_t pos)
{
    if (pos != cursor) {
        cursor = pos;
        sig_task_changed.raise(false);
    }
}

/** Get cursor. */
uint32_t
WAutoTaskObjectSelection::getCursor() const
{
    return cursor;
}

/** Get associated notification message. */
IntNotificationMessage*
WAutoTaskObjectSelection::getNotificationMessage() const
{
    return current_process != 0 ? current_process->getNotificationMessage() : 0;
}

/** Get editor. */
const IntAutoTaskEditor&
WAutoTaskObjectSelection::getEditor() const
{
    return editor;
}

/** Replace task text. \see IntAutoTaskEditor::replace */
void
WAutoTaskObjectSelection::replace(uint32_t pos, uint32_t nold, const string_t& line)
{
    replace(pos, nold, &line, 1);
}

/** Replace task text. \see IntAutoTaskEditor::replace */
void
WAutoTaskObjectSelection::replace(uint32_t pos, uint32_t nold, const string_t* line, uint32_t nnew)
{
    // Update task
    editor.replace(pos, nold, line, nnew);

    // Update cursor. We have the following cases:
    if (cursor < pos) {
        // cursor is before change; no modification
    } else if (cursor < pos+nold) {
        // cursor is within changed area: move to beginning of change
        cursor = pos;
    } else {
        // cursor is after changed area
        cursor -= nold;
        cursor += nnew;
    }

    // Update observers
    sig_task_changed.raise(false);
}

/** Insert one command. */
void
WAutoTaskObjectSelection::insert(const string_t& line)
{
    replace(getCursor(), 0, &line, 1);
}

/** Set program counter. */
void
WAutoTaskObjectSelection::setPC()
{
    editor.setPC(cursor);
    sig_task_changed.raise(false);
}

/** Common command GUI.
    \param sender Sender widget, for meun
    \param allowName Allow "SetName" command */
void
WAutoTaskObjectSelection::createCommonCommand(UIBaseWidget& sender, bool allowName)
{
    // Create list
    StringList sl;
    sl.add(1, _("Notify message"));
    sl.add(2, _("Set Comment"));
    if (allowName) {
        sl.add(3, _("Set Name"));
    }
    sl.add(4, _("Wait One Turn"));
    sl.add(5, _("Restart task from beginning"));

    // Menu
    UIStandardListbox list(0, sl.getCount(), getTextWidth(*fonts[FONT_NORMAL], sl) + 20, sl);
    if (!list.doStandardMenu(sender)) {
        return;
    }

    // Act
    string_t val;
    switch (list.getKey()) {
     case 1:
        if (UIInputLine::run(val, 255, _("Notify"), _("Enter message:"), 30, 0)) {
            insert("Notify " + quoteString(val));
        }
        break;
     case 2:
        if (UIInputLine::run(val, 255, _("Set Comment"), _("Enter new comment:"), 30, UIInputLine::ilf_GameChars)) {
            insert("SetComment " + quoteString(val));
        }
        break;
     case 3:
        if (UIInputLine::run(val, 20, _("Set Name"), _("Enter new name:"), 20, UIInputLine::ilf_GameChars)) {
            insert("SetName " + quoteString(val));
        }
        break;
     case 4:
        insert("WaitOneTurn");
        break;
     case 5:
        insert("Restart");
        break;
    }
}

/** Called when a different object is selected. */
void
WAutoTaskObjectSelection::onCurrentChanged()
{
    loadTask(getTask(true));
    sig_task_changed.raise(true);
}

/** Called when the object we're observing changes. */
void
WAutoTaskObjectSelection::onObjectChanged()
{
    if (current_process != 0) {
        sig_task_changed.raise(false);
    }
}

/** Load auto task.
    \param process Auto task process. Can be 0 to unload current process. */
void
WAutoTaskObjectSelection::loadTask(IntExecutionContext* process)
{
    /* Quick out */
    if (process == current_process) {
        return;
    }

    /* Flush original process and make it runnable */
    IntExecutionContext* old_process = current_process;
    if (current_process != 0) {
        editor.saveToProcess(*current_process);
        current_process->setState(IntExecutionContext::psRunnable);
    }

    /* Load new process, and freeze it */
    current_process = process;
    if (current_process != 0) {
        editor.loadFromProcess(*current_process);
        current_process->setState(IntExecutionContext::psFrozen);
    } else {
        editor.clear();
    }

    /* Set cursor */
    cursor = editor.getNumInstructions();

    /* Run old process */
    if (old_process != 0 && old_process->getState() == IntExecutionContext::psRunnable) {
        old_process->run();
    }
}


/**************************** WAutoTaskScreen ****************************/

/** Auto task screen.
    \param selection Object selection
    \param screenId Screen identifier */
WAutoTaskScreen::WAutoTaskScreen(WAutoTaskObjectSelection& selection, WScreenId::ScreenId screenId)
    : WControlScreen(selection, screenId),
      selection(selection)
{ }

WAutoTaskScreen::~WAutoTaskScreen()
{ }

bool
WAutoTaskScreen::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        if (event.key.code == SDLK_DELETE) {
            // Delete
            selection.replace(selection.getCursor(), 1, 0, 0);
            return true;
        } else if (event.key.code == SDLK_INSERT) {
            // Insert
            string_t cmd;
            editCommand(cmd, _("Insert Command"), selection, 0);
            return true;
        } else if (event.key.code == ' ') {
            // Edit
            string_t cmd;
            if (selection.getCursor() < selection.getEditor().getNumInstructions()) {
                cmd = selection.getEditor()[selection.getCursor()];
            }
            editCommand(cmd, _("Edit Command"), selection, 1);
            return true;
        } else if (event.key.code == '%' || event.key.code == 'k') {
            // Comment in/out
            if (selection.getCursor() < selection.getEditor().getNumInstructions()) {
                string_t cmd = selection.getEditor()[selection.getCursor()];
                if (cmd.size() > 0 && cmd[0] == '%') {
                    cmd.erase(0, cmd.find_first_not_of(" %", 0));
                } else {
                    cmd.insert(0, "% ");
                }
                if (IntAutoTaskEditor::isValidCommand(cmd)) {
                    selection.replace(selection.getCursor(), 1, cmd);
                }
            }
            return true;
        } else if (event.key.code == ss_Ctrl + SDLK_DELETE) {
            // Delete all
            if (UIMessageBox(_("Delete this auto task?"), _("Auto Task")).doYesNoDialog()) {
                selection.replace(0, selection.getEditor().getNumInstructions(), 0, 0);
            }
            return true;
        } else if (event.key.code == ss_Ctrl + 'n') {
            // Set PC
            selection.setPC();
            return true;
        } else if (event.key.code == ss_Ctrl + 'r') {
            // read file
            handleLoad();
            return true;
        } else if (event.key.code == ss_Ctrl + 's') {
            // save file
            handleSave();
            return true;
        }
    }
    return WControlScreen::handleEvent(event, second_pass);
}

/** "Save" command. */
void
WAutoTaskScreen::handleSave()
{
    // Ask for file
    UIFileChooserDialog fc(_("Save Task"), Directory::openDir(Directory::getWorkingDirectory()), "*.cct");
    ui_root->centerWidget(fc);
    if (!ui_root->runModal(fc)) {
        return;
    }

    // Save
    try {
        Ptr<Stream> out = Stream::openFile(fc.getResultFile(), Stream::C_CREATE);
        TextFile tf(*out);
        tf.setUtf8();
        for (uint32_t i = 0, e = selection.getEditor().getNumInstructions(); i < e; ++i) {
            tf.writeLine(selection.getEditor()[i]);
        }
        tf.sync();
    }
    catch (FileProblemException& e) {
        messageBox(format(_("Unable to create file %s: %s"), e.getFileName(), e.what()), _("Save Task"));
    }
    catch (std::exception& e) {
        messageBox(format(_("Unable to create file %s: %s"), fc.getResultFile(), e.what()), _("Save Task"));
    }
}

/** "Load" command. */
void
WAutoTaskScreen::handleLoad()
{
    // Ask for file
    UIFileChooserDialog fc(_("Load Task"), Directory::openDir(Directory::getWorkingDirectory()), "*.cct");
    ui_root->centerWidget(fc);
    if (!ui_root->runModal(fc)) {
        return;
    }

    // Save
    try {
        Ptr<Stream> out = Stream::openFile(fc.getResultFile(), Stream::C_READ);
        TextFile tf(*out);
        tf.setUtf8();
        string_t line;
        int numBadLines = 0;
        while (tf.getLine(line)) {
            if (IntAutoTaskEditor::isValidCommand(line)) {
                selection.insert(line);
            } else {
                ++numBadLines;
            }
        }
        if (numBadLines != 0) {
            messageBox(format(_("This file contained %d line%1{ which is%|s which are%} not permitted in Auto Tasks. "
                                "%1{It has%|They have%} been skipped."), numBadLines),
                       _("Load Task"));
        }
    }
    catch (FileProblemException& e) {
        messageBox(format(_("Unable to open file %s: %s"), e.getFileName(), e.what()), _("Load Task"));
    }
    catch (std::exception& e) {
        messageBox(format(_("Unable to open file %s: %s"), fc.getResultFile(), e.what()), _("Load Task"));
    }
}
