/**
  *  \file client/widgets/markertype.cc
  *  \brief WMarkerTypeSelector widget
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/markertype.h"
#include "client/marks.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "ui/widgets/statictext.h"
#include "ui/window.h"
#include "util/nls.h"
#include "ui/widgets/quit.h"

/************************** WMarkerTypeSelector **************************/

WMarkerTypeSelector::WMarkerTypeSelector()
    : marker_type(0)
{ }

WMarkerTypeSelector::~WMarkerTypeSelector()
{ }

/** Move cursor by specified delta. */
void
WMarkerTypeSelector::moveBy(int delta)
{
    int n = marker_type + delta;
    if (n <= 0)
        n += NUM_USER_MARKERS;
    if (n > NUM_USER_MARKERS)
        n -= NUM_USER_MARKERS;
    setMarkerType(n);
}

/** Get size of one color box. */
int
WMarkerTypeSelector::getFrameSize() const
{
    return font_heights[FONT_TITLE];
}

void
WMarkerTypeSelector::setMarkerType(int marker_type)
{
    if (marker_type >= 0 && marker_type < NUM_USER_MARKERS) {
        if (marker_type != this->marker_type) {
            this->marker_type = marker_type;
            drawWidget();
        }
    }
}
int
WMarkerTypeSelector::getMarkerType() const
{
    return marker_type;
}

void
WMarkerTypeSelector::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = GfxPoint(NUM_USER_MARKERS*getFrameSize(), getFrameSize());
}

void
WMarkerTypeSelector::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.setColor(UIColor::tc_Static);

    const int fs = getFrameSize();
    for (int col = 0; col < NUM_USER_MARKERS; ++col) {
        const int x = getExtent().x + col*fs;
        const int y = getExtent().y;
        // Draw frame
        if (col == marker_type) {
            drawFrameDown(can, GfxRect(x, y, fs, fs));
        } else {
            drawFrameUp(can, GfxRect(x, y, fs, fs));
        }

        // Draw box
        drawBackground(ctx, GfxRect(x+1, y+1, fs-2, fs-2));

        // Draw marker
        drawMarker(ctx, getUserMarker(col, true), GfxPoint(x + fs/2, y + fs/2));
    }
    can.update(getExtent());
}

bool
WMarkerTypeSelector::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Mouse && e.mouse.b != 0 && isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
        // Mouse clicked inside widget
        setSelectedWidget();
        int col  = (e.mouse.x - getExtent().x) / getFrameSize();
        if (col < NUM_USER_MARKERS)
            setMarkerType(col);
        return true;
    }
    if (e.type == ev_Key) {
        // Handle some keys
        switch (e.key.code) {
         case SDLK_LEFT:
            setSelectedWidget();
            moveBy(-1);
            return true;
         case SDLK_RIGHT:
            setSelectedWidget();
            moveBy(+1);
            return true;
        }
    }
    return UIWidget::handleEvent(e, second_pass);
}

bool
WMarkerTypeSelector::doStandardDialog(string_t title)
{
    ObjectHolder h;
    UIWindow w(title, UIWindow::BLUE_WINDOW, COLOR_GRAY);
    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    w.setLayout(UIVBoxLayout::instance5);
    w.add(h.add(new UIStaticText(_("Choose marker type:"), UIStaticText::StaticTitle)));
    w.add(*this);
    g.add(h.add(new UIButton(true,  _("OK"), SDLK_RETURN)).addStop());
    g.add(h.add(new UIButton(false, _("Cancel"), SDLK_ESCAPE)).addStop());
    g.add(h.add(new UISpacer()));
    w.add(g);
    w.add(h.add(new UIQuit(false)));
    w.pack();
    ui_root->centerWidget(w);
    return ui_root->runModal(w);
}

