/**
  *  \file game/coord.h
  *  \brief Game Coordinate System Handling
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_COORD_H
#define PCC_V2_GAME_COORD_H

#include <iosfwd>
#include "cpluslib/string.h"

//! Point in Game Universe.
struct GPoint {
    int  x, y;
    GPoint(int ax, int ay)
        : x(ax), y(ay) { }
    GPoint() { }

    bool   isOnMap() const;

    GPoint  getCanonicalLocation() const;
    GPoint  getSimpleCanonicalLocation() const;
    GPoint  getSimpleNearestAlias(GPoint a) const;

    bool    operator==(GPoint rhs)   const { return x == rhs.x && y == rhs.y; }
    bool    operator!=(GPoint rhs)   const { return x != rhs.x || y != rhs.y; }
    GPoint  operator+ (GPoint other) const { return GPoint(x + other.x, y + other.y); }
    GPoint  operator- (GPoint other) const { return GPoint(x - other.x, y - other.y); }
    GPoint& operator+=(GPoint other)       { x += other.x; y += other.y; return *this; }
    GPoint& operator-=(GPoint other)       { x -= other.x; y -= other.y; return *this; }

    bool parseCoordinates(string_t s);
    bool parseSectorNumber(const string_t& s);
    bool parseSectorNumber(int n);

    int  getSectorNumber() const;
};

/** Starchart configuration. */
class GChartConfiguration {
 public:
    GChartConfiguration();

    enum Mode {
        Flat,
        Wrapped,
        Circular
    };

    // Configuration inquiry
    Mode getMode() const;
    GPoint getCenter() const;
    GPoint getSize() const;
    GPoint getMinimumCoordinates() const;
    GPoint getMaximumCoordinates() const;

    // Configuration
    void initFromConfig();
    void setConfig(Mode mode, GPoint center, GPoint size);
    bool isSetFromGameConfig() const;

    // Coordinate management
    bool isOnMap(const GPoint& pt) const;
    bool isValidPlanetCoordinate(const GPoint& pt) const;
    GPoint getCanonicalLocation(GPoint pt) const;
    GPoint getSimpleCanonicalLocation(GPoint pt) const;
    GPoint getSimpleNearestAlias(GPoint pt, GPoint a) const;

    int  getNumRectangularImages() const;
    int  getNumPointImages() const;
    bool getPointAlias(GPoint pt, GPoint& out, int image, bool exact) const;
    GPoint getSimplePointAlias(GPoint pt, int image) const;

 private:
    void computeDerivedInformation();

    Mode    mode;
    GPoint  center;
    GPoint  size;
    GPoint  min;
    GPoint  max;
    bool    from_hconfig;
};


extern GChartConfiguration chart_conf;

long distanceSquaredRaw(GPoint a, GPoint b);
long distanceSquared(GPoint a, GPoint b);

std::ostream& operator<<(std::ostream&, GPoint pt);

extern const struct ModuleInitializer mod_coord[];


/******************************** Inlines ********************************/

/** Check if point is on map. */
inline bool
GPoint::isOnMap() const
{
    return chart_conf.isOnMap(*this);
}

inline GPoint
GPoint::getCanonicalLocation() const
{
    return chart_conf.getCanonicalLocation(*this);
}

inline GPoint
GPoint::getSimpleCanonicalLocation() const
{
    return chart_conf.getSimpleCanonicalLocation(*this);
}

inline GPoint
GPoint::getSimpleNearestAlias(GPoint a) const
{
    return chart_conf.getSimpleNearestAlias(*this, a);
}

inline bool
GChartConfiguration::isSetFromGameConfig() const
{
    return from_hconfig;
}

inline GChartConfiguration::Mode
GChartConfiguration::getMode() const
{
    return mode;
}

inline GPoint
GChartConfiguration::getCenter() const
{
    return center;
}

inline GPoint
GChartConfiguration::getSize() const
{
    return size;
}

inline GPoint
GChartConfiguration::getMinimumCoordinates() const
{
    return min;
}

inline GPoint
GChartConfiguration::getMaximumCoordinates() const
{
    return max;
}
#endif
