/**
  *  \file game/load.h
  *  \brief Loading Functions
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_LOAD_H
#define PCC_V2_GAME_LOAD_H

#include "cpluslib/smartptr.h"
#include "cpluslib/types.h"
#include "game/struct.h"
#include "cpluslib/assert.h"
#include "game/playerset.h"

class GGameTurn;
class Stream;
class Directory;

/** \defgroup game_load Game Loading Functions.
    These functions load game data into a GGameTurn. @{ */

/** Target for a Load operation. Note that these must have these exact
    values, as the loadXXX functions will do 'where & LoadCurrent'. */
enum LoadUniverse {
    LoadCurrent  = 1,
    LoadPrevious = 2,
    LoadBoth     = LoadCurrent | LoadPrevious
};

/** Target file format. */
enum TargetFormat {
    TargetPlaintext,            ///< Plaintext file. Standard in Dosplan etc.
    TargetEncrypted             ///< Encrypted file. Winplan's additional targets.
};

/** Logic to read an outbox file. */
class GOutboxReader {
 public:
    /** Add message read from file.
        \param message Message text, decoded
        \param receivers Receiver set */
    virtual void addMessage(string_t message, GPlayerSet receivers) = 0;

    void loadOutbox(Stream& s);
    void loadOutbox35(Stream& s);
};

/* single-load routines, load.cc */
void loadShips(GGameTurn& trn, Stream& s, int count, LoadUniverse where, GPlayerSet src);
void loadPlanets(GGameTurn& trn, Stream& s, int count, LoadUniverse where, GPlayerSet src);
void loadPlanetXY(GGameTurn& trn, Stream& s, LoadUniverse where);
void loadPlanetNames(GGameTurn& trn, Stream& s, LoadUniverse where);
void loadBases(GGameTurn& trn, Stream& s, int count, LoadUniverse where, GPlayerSet src);
void loadTargets(GGameTurn& trn, Stream& s, int count, TargetFormat fmt, LoadUniverse where, GPlayerSet src);
void loadShipXY(GGameTurn& trn, Stream& s, int avail, LoadUniverse where, GPlayerSet src, GPlayerSet reject);
void loadKoreMinefields(GGameTurn& trn, Stream& s, int count, int player);
void loadKoreIonStorms(GGameTurn& trn, Stream& s, int count);
void loadStormNames(GGameTurn& trn, Stream& s);
void loadUfos(GGameTurn& trn, Stream& s, int first_id, int count);
void loadKoreExplosions(GGameTurn& trn, Stream& s, int count);
void addMessageFromFile(GGameTurn& trn, string_t text, int sender, GPlayerSet recv);
void maybeLoadFlakVcrs(GGameTurn& trn, Directory& dir, int player);

void loadCommon(GGameTurn& trn, Directory& gameDir, Directory& specDir, LoadUniverse where, int player);

/* complete routines, load-{rst,trn,dir}.cc */
void loadResult(GGameTurn& trn, Stream& s, int player, bool withBackup);
void loadTurn(GGameTurn& trn, Stream& s, int player);
void loadDirectory(GGameTurn& trn, Directory& d, int player);
void loadHostDirectory(GGameTurn& trn, Directory& d);

/*@}*/

#endif
