/**
  *  \file game/objlist.cc
  *  \brief Object Lists
  *
  *  This module contains the simple parts of object list management.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/objlist.h"
#include "cpluslib/assert.h"
#include "game/global.h"
#include "game/ship.h"
#include "game/planet.h"

/**************************** GObjectReference ***************************/

GObjectReference::GObjectReference()
    : type(0),
      index(0)
{ }

GObjectReference::GObjectReference(GObjectType& type, int index)
{
    if (GObjectList* list = dynamic_cast<GObjectList*>(&type)) {
        // Dereference link
        *this = (*list)[index];
    } else {
        // Reference the object
        this->type = &type;
        this->index = index;
    }
}

/****************************** GObjectList *****************************/

GObjectList::GObjectList()
    : list(),
      notes()
{ }

GObjectList::~GObjectList()
{ }

// Adding things:
/** Add a single object.
    \param ref Reference to object
    \pre ref.isValid() */
void
GObjectList::addObject(GObjectReference ref)
{
    ASSERT(ref.isValid());
    list.push_back(ref);
}

/** Add a single object.
    \param type Type of object
    \param index Index into type */
void
GObjectList::addObject(GObjectType& type, int index)
{
    list.push_back(GObjectReference(type, index));
}

/** Add all objects at a particular location. This is used for certain ship lists.
    \param univ  universe
    \param pt    location
    \param flags ls_IncludePlanet: include planet in list;
                 ls_IncludeForeignShips: also list foreign ships
                 (default: only list ships we're playing);
                 ls_SafeShipsOnly: include only safe ships (not guessed)
    \param exclude_ship do not include this ship

    \todo We need a means to say which player must safely see the ship */
void
GObjectList::addObjectsAt(GUniverse& univ, GPoint pt, int flags, int exclude_ship)
{
    // Canonical location
    pt = pt.getCanonicalLocation();

    // Handle planet
    if ((flags & ls_IncludePlanet) != 0) {
        if (int pid = univ.getPlanetAt(pt)) {
            addObject(univ.ty_any_planets, pid);
        }
    }

    // Handle ships
    for (int sid = univ.ty_any_ships.findNextIndex(0); sid != 0; sid = univ.ty_any_ships.findNextIndex(sid)) {
        GShip& s = univ.getShip(sid);
        if (sid != exclude_ship
            && s.getPos() == pt
            && ((flags & ls_IncludeForeignShips) || s.isPlayable(GObject::Playable))
            && (!(flags & ls_SafeShipsOnly) || s.isReliablyVisible(0)))
        {
            addObject(univ.ty_any_ships, sid);
        }
    }
}

/** Add non-object.
    \param id Id number. 0=divider, otherwise regular non-object
    \param name Name */
void
GObjectList::addNonObject(int id, string_t name)
{
    addObject(notes, notes.add(id, name));
}

// Manipulation:
/** Clear list. */
void
GObjectList::clear()
{
    container_t().swap(list);
    notes.clear();
    sig_set_changed.raise(0);
}

/** Add dividers according a function. The passed function will return
    a string for every object; a heading (divider) will be added whenever
    that string changes. For example, when the list contains
    \verbatim
       ...object...    ...func returns...
       Ship #1         Crystalline
       Ship #9         Crystalline
       Ship #3         Empire
       Ship #17        Crystalline
    \endverbatim
    the list will look like this afterwards ("==foo==" are dividers):
    \verbatim
       == Crystalline ==
       Ship #1
       Ship #9
       == Empire ==
       Ship #3
       == Crystalline ==
       Ship #17
    \endverbatim */
void
GObjectList::addDividers(string_t func(GObject&))
{
    class Helper : public GObjectDivider {
     public:
        Helper(string_t func(GObject&))
            : func(func)
            { }
        string_t getDivider(GObject& obj)
            { return func(obj); }
     private:
        string_t (*func)(GObject&);
    };
    Helper h(func);
    addDividers(h);
}

/** Add dividers according an object.
    \overload */
void
GObjectList::addDividers(GObjectDivider& func)
{
    // Room for new data
    container_t    tmpList;
    GNoteContainer tmpNotes;

    // Current divider
    string_t current;

    // Build
    for (container_t::iterator i = list.begin(); i != list.end(); ++i) {
        if (i->isGameObject()) {
            // Game object with possible divider
            string_t next = func.getDivider(**i);
            if (next.length() && current != next) {
                current = next;
                tmpList.push_back(GObjectReference(notes, tmpNotes.add(0, current)));
            }
            tmpList.push_back(*i);
        } else if (i->isValid()) {
            // Not a game object, might be a non-divider non-object
            GNoteObject* note = dynamic_cast<GNoteObject*>(&**i);
            if (note != 0 && note->getId() != 0) {
                tmpList.push_back(GObjectReference(notes, tmpNotes.addCopyOf(*note)));
            }
        }
    }
    tmpList.swap(list);
    tmpNotes.swap(notes);
    sig_set_changed.raise(0);
}

void
GObjectList::removeDividers()
{
    /* in-place deletion */
    container_t::iterator o = list.begin();
    for (container_t::iterator i = o; i != list.end(); ++i) {
        if (i->isGameObject() || (*i)->getId() != 0) {
            *o = *i;
            ++o;
        }
    }
    list.erase(o, list.end());
    sig_set_changed.raise(0);
}

// Container access: (0-based)
GObjectList::iterator
GObjectList::begin() const
{
    return list.begin();
}

GObjectList::iterator
GObjectList::end() const
{
    return list.end();
}

GObjectList::size_type
GObjectList::size() const
{
    return list.size();
}

const GObjectReference&
GObjectList::operator[](size_type index) const
{
    return list[index];
}

// GObjectType: (1-based!)
GObject&
GObjectList::getObjectByIndex(int index)
{
    return *list[index-1];
}

GUniverse*
GObjectList::getUniverseByIndex(int index)
{
    return list[index-1].getUniverse();
}

bool
GObjectList::isValidIndex(int index) const
{
    return index > 0
        && index <= int(list.size())
        && list[index-1].getObjectType().isValidIndex(list[index-1].getObjectIndex());
}

int
GObjectList::getNextIndex(int index) const
{
    return index < int(list.size()) ? index+1 : 0;
}

int
GObjectList::getPreviousIndex(int index) const
{
    return index > 0 ? index-1 : list.size();
}

const GObjectReference&
GObjectList::getObjectReferenceByIndex(int index) const
{
    return list[index-1];
}

int
GObjectList::getIndexFor(GObjectReference id) const
{
    for (size_type i = 0; i < list.size(); ++i) {
        if (list[i] == id) {
            return i+1;
        }
    }
    return 0;
}

int
GObjectList::getIndexFor(const GObject& obj) const
{
    for (size_type i = 0; i < list.size(); ++i) {
        if (&*list[i] == &obj) {
            return i+1;
        }
    }
    return 0;
}

GObjectList::size_type
GObjectList::countGameObjects() const
{
    size_type count = 0;
    for (size_type i = 0; i < list.size(); ++i) {
        if (list[i].isGameObject()) {
            ++count;
        }
    }
    return count;
}
