/**
  *  \file u/t_game_cost.cc
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "u/t_game.h"
#include "game/cargo.h"
#include "game/cost.h"

/** Tests various cases of GCost::fromString.

    This does not test invalid cases, as those are not yet defined. As of 20081216,
    invalid characters are ignored by GCost::fromString, and treated as if not present.
    Possible cases that need definition:
    - invalid cargo letters
    - overflow
    - signs */
void
GameCostTestSuite::testParse()
{
    /* Like PCC 1.x, we accept cargospecs and PHost format:
          Cargospec:  123TDM
          PHost:      T123 D123 M123 */
    {
        // Zero-initalisation
        GCost value;
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 0);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(value.get(el_Money), 0);
        TS_ASSERT_EQUALS(value.toPHostString(), "S0");
    }

    {
        // Blank cargospec
        GCost value = GCost::fromString("");
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 0);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(value.get(el_Money), 0);
        TS_ASSERT_EQUALS(value.toPHostString(), "S0");
    }

    {
        // Zero cargospec
        GCost value = GCost::fromString("0td");
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 0);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(value.get(el_Money), 0);
        TS_ASSERT_EQUALS(value.toPHostString(), "S0");
    }

    {
        // Standard cargospec (torpedo cost)
        GCost value = GCost::fromString("1tdm 20$");
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 1);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 1);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 1);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(value.get(el_Money), 20);
        TS_ASSERT_EQUALS(value.toPHostString(), "T1 D1 M1 $20");
    }

    {
        // Standard cargospec without space
        GCost value = GCost::fromString("1tdm42$");
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 1);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 1);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 1);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(value.get(el_Money), 42);
        TS_ASSERT_EQUALS(value.toPHostString(), "T1 D1 M1 $42");
    }

    {
        // Standard cargospec with duplication
        GCost value = GCost::fromString("1ttttdm");
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 4);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 1);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 1);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(value.get(el_Money), 0);
        TS_ASSERT_EQUALS(value.toPHostString(), "T4 D1 M1");
    }

    {
        // Standard cargospec with addition
        GCost value = GCost::fromString("10s 20s");
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 0);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 30);
        TS_ASSERT_EQUALS(value.get(el_Money), 0);
        TS_ASSERT_EQUALS(value.toPHostString(), "S30");
    }

    {
        // Standard cargospec, uppercase
        GCost value = GCost::fromString("10TDM 99S");
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 10);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 10);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 10);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 99);
        TS_ASSERT_EQUALS(value.get(el_Money), 0);
        TS_ASSERT_EQUALS(value.toPHostString(), "T10 D10 M10 S99");
    }

    {
        // PHost-style
        GCost value = GCost::fromString("T10 D20 M30 $77 S42");
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 10);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 20);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 30);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 42);
        TS_ASSERT_EQUALS(value.get(el_Money), 77);
        TS_ASSERT_EQUALS(value.toPHostString(), "T10 D20 M30 S42 $77");
    }

    {
        // PHost-style, lower-case
        GCost value = GCost::fromString("t11 d22 m33 $44 S55");
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 11);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 22);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 33);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 55);
        TS_ASSERT_EQUALS(value.get(el_Money), 44);
        TS_ASSERT_EQUALS(value.toPHostString(), "T11 D22 M33 S55 $44");
    }

    {
        // PHost-style, with addition
        GCost value = GCost::fromString("t11 t22 t33");
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 66);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(value.get(el_Money), 0);
        TS_ASSERT_EQUALS(value.toPHostString(), "T66");
    }

    {
        // Initialisation from TCost structure
        const TCost v = { 5, 6, 7, 8 }; // mc, t, d, m
        GCost value = v;
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 6);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 7);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 8);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(value.get(el_Money), 5);
        TS_ASSERT_EQUALS(value.toPHostString(), "T6 D7 M8 $5");
    }

    {
        // GCost parses using GCargoSpec::parse(..., true), so check "max" syntax
        GCost value = GCost::fromString("tmax");
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 10000);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(value.get(el_Money), 0);
    }

    {
        GCost value = GCost::fromString("tm");
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 10000);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(value.get(el_Money), 0);
    }

    {
        GCost value = GCost::fromString("tmax d10");
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 10000);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 10);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(value.get(el_Money), 0);
    }

    {
        GCost value = GCost::fromString("tm d10");
        TS_ASSERT_EQUALS(value.get(el_Tritanium), 10000);
        TS_ASSERT_EQUALS(value.get(el_Duranium), 10);
        TS_ASSERT_EQUALS(value.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(value.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(value.get(el_Money), 0);
    }
}

void
GameCostTestSuite::testAdd()
{
    // +=
    {
        GCost a = GCost::fromString("t1");
        GCost b = GCost::fromString("t42");
        a += b;
        TS_ASSERT_EQUALS(a.get(el_Tritanium), 43);
        TS_ASSERT_EQUALS(a.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(a.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(a.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(a.get(el_Money), 0);
        TS_ASSERT(a.isValid());
        TS_ASSERT(b.isValid());
    }

    {
        GCost a = GCost::fromString("t1");
        GCost b = GCost::fromString("s42");
        a += b;
        TS_ASSERT_EQUALS(a.get(el_Tritanium), 1);
        TS_ASSERT_EQUALS(a.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(a.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(a.get(el_Supplies), 42);
        TS_ASSERT_EQUALS(a.get(el_Money), 0);
        TS_ASSERT(a.isValid());
        TS_ASSERT(b.isValid());
    }

    {
        GCost a = GCost::fromString("s100");
        GCost b = GCost::fromString("$200");
        a += b;
        TS_ASSERT_EQUALS(a.get(el_Tritanium), 0);
        TS_ASSERT_EQUALS(a.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(a.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(a.get(el_Supplies), 100);
        TS_ASSERT_EQUALS(a.get(el_Money), 200);
        TS_ASSERT(a.isValid());
        TS_ASSERT(b.isValid());
    }

    {
        GCost a;
        GCost b = GCost::fromString("$200");
        a += b;
        TS_ASSERT_EQUALS(a.get(el_Tritanium), 0);
        TS_ASSERT_EQUALS(a.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(a.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(a.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(a.get(el_Money), 200);
        TS_ASSERT(a.isValid());
        TS_ASSERT(b.isValid());
    }
}

void
GameCostTestSuite::testSubtract()
{
    // -=
    {
        GCost a = GCost::fromString("t1");
        GCost b = GCost::fromString("t42");
        a -= b;
        TS_ASSERT_EQUALS(a.get(el_Tritanium), -41);
        TS_ASSERT_EQUALS(a.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(a.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(a.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(a.get(el_Money), 0);
        TS_ASSERT(!a.isValid());
        TS_ASSERT(b.isValid());
    }

    {
        GCost a = GCost::fromString("t1");
        GCost b = GCost::fromString("s42");
        a -= b;
        TS_ASSERT_EQUALS(a.get(el_Tritanium), 1);
        TS_ASSERT_EQUALS(a.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(a.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(a.get(el_Supplies), -42);
        TS_ASSERT_EQUALS(a.get(el_Money), 0);
        TS_ASSERT(!a.isValid());
        TS_ASSERT(b.isValid());
    }

    {
        GCost a = GCost::fromString("s100");
        GCost b = GCost::fromString("$200");
        a -= b;
        TS_ASSERT_EQUALS(a.get(el_Tritanium), 0);
        TS_ASSERT_EQUALS(a.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(a.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(a.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(a.get(el_Money), -100);
        TS_ASSERT(!a.isValid());
        TS_ASSERT(b.isValid());
    }

    {
        GCost a = GCost::fromString("$200");
        GCost b;
        a -= b;
        TS_ASSERT_EQUALS(a.get(el_Tritanium), 0);
        TS_ASSERT_EQUALS(a.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(a.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(a.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(a.get(el_Money), 200);
        TS_ASSERT(a.isValid());
        TS_ASSERT(b.isValid());
    }
}

void
GameCostTestSuite::testMult()
{
    // *=, *
    {
        GCost a;
        a *= 10;
        TS_ASSERT_EQUALS(a.get(el_Tritanium), 0);
        TS_ASSERT_EQUALS(a.get(el_Duranium), 0);
        TS_ASSERT_EQUALS(a.get(el_Molybdenum), 0);
        TS_ASSERT_EQUALS(a.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(a.get(el_Money), 0);
    }

    {
        GCost a = GCost::fromString("3tdm 42$");
        a *= 10;
        TS_ASSERT_EQUALS(a.get(el_Tritanium), 30);
        TS_ASSERT_EQUALS(a.get(el_Duranium), 30);
        TS_ASSERT_EQUALS(a.get(el_Molybdenum), 30);
        TS_ASSERT_EQUALS(a.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(a.get(el_Money), 420);
    }

    {
        GCost a = GCost::fromString("3tdm 42$");
        GCost b = a * 10;
        TS_ASSERT_EQUALS(a.get(el_Tritanium), 3);
        TS_ASSERT_EQUALS(a.get(el_Duranium), 3);
        TS_ASSERT_EQUALS(a.get(el_Molybdenum), 3);
        TS_ASSERT_EQUALS(a.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(a.get(el_Money), 42);
        TS_ASSERT_EQUALS(b.get(el_Tritanium), 30);
        TS_ASSERT_EQUALS(b.get(el_Duranium), 30);
        TS_ASSERT_EQUALS(b.get(el_Molybdenum), 30);
        TS_ASSERT_EQUALS(b.get(el_Supplies), 0);
        TS_ASSERT_EQUALS(b.get(el_Money), 420);
    }

    {
        GCost a = GCost::fromString("1t 2d 3m 4$ 5s") * 7;
        TS_ASSERT_EQUALS(a.get(el_Tritanium), 7);
        TS_ASSERT_EQUALS(a.get(el_Duranium), 14);
        TS_ASSERT_EQUALS(a.get(el_Molybdenum), 21);
        TS_ASSERT_EQUALS(a.get(el_Supplies), 35);
        TS_ASSERT_EQUALS(a.get(el_Money), 28);
    }
}

void
GameCostTestSuite::testCompare()
{
    // ==, !=
    TS_ASSERT(GCost::fromString("") == GCost());
    TS_ASSERT(GCost::fromString("100$") == GCost::fromString("$100"));
    TS_ASSERT(GCost::fromString("5tdm") == GCost::fromString("T5 5M 5d"));
    TS_ASSERT(GCost::fromString("5tdm") != GCost::fromString("T5 5M 5d 1d"));
    TS_ASSERT(GCost::fromString("1t") != GCost());
    TS_ASSERT(GCost::fromString("1d") != GCost());
    TS_ASSERT(GCost::fromString("1m") != GCost());
    TS_ASSERT(GCost::fromString("1$") != GCost());
    TS_ASSERT(GCost::fromString("1s") != GCost());
    TS_ASSERT(GCost::fromString("t1") != GCost());
    TS_ASSERT(GCost::fromString("d1") != GCost());
    TS_ASSERT(GCost::fromString("m1") != GCost());
    TS_ASSERT(GCost::fromString("$1") != GCost());
    TS_ASSERT(GCost::fromString("s1") != GCost());
    TS_ASSERT(GCost::fromString("s100") != GCost::fromString("$100"));
    TS_ASSERT(GCost::fromString("$100") != GCost::fromString("s100"));
}

void
GameCostTestSuite::testEnough()
{
    // isEnoughFor()
    // Equality:
    TS_ASSERT( GCost().isEnoughFor(GCost()));
    TS_ASSERT( GCost::fromString("1t").isEnoughFor(GCost::fromString("1t")));
    TS_ASSERT( GCost::fromString("1d").isEnoughFor(GCost::fromString("1d")));
    TS_ASSERT( GCost::fromString("1m").isEnoughFor(GCost::fromString("1m")));
    TS_ASSERT( GCost::fromString("1s").isEnoughFor(GCost::fromString("1s")));
    TS_ASSERT( GCost::fromString("1$").isEnoughFor(GCost::fromString("1$")));

    // Strictly more:
    TS_ASSERT(!GCost::fromString("1t").isEnoughFor(GCost::fromString("2t")));
    TS_ASSERT(!GCost::fromString("1d").isEnoughFor(GCost::fromString("2d")));
    TS_ASSERT(!GCost::fromString("1m").isEnoughFor(GCost::fromString("2m")));
    TS_ASSERT(!GCost::fromString("1s").isEnoughFor(GCost::fromString("2s")));
    TS_ASSERT(!GCost::fromString("1$").isEnoughFor(GCost::fromString("2$")));

    // Element (Non-)Conversions:
    TS_ASSERT( GCost::fromString("1t").isEnoughFor(GCost::fromString("1t")));
    TS_ASSERT(!GCost::fromString("1d").isEnoughFor(GCost::fromString("1t")));
    TS_ASSERT(!GCost::fromString("1m").isEnoughFor(GCost::fromString("1t")));
    TS_ASSERT(!GCost::fromString("1s").isEnoughFor(GCost::fromString("1t")));
    TS_ASSERT(!GCost::fromString("1$").isEnoughFor(GCost::fromString("1t")));

    TS_ASSERT(!GCost::fromString("1t").isEnoughFor(GCost::fromString("1d")));
    TS_ASSERT( GCost::fromString("1d").isEnoughFor(GCost::fromString("1d")));
    TS_ASSERT(!GCost::fromString("1m").isEnoughFor(GCost::fromString("1d")));
    TS_ASSERT(!GCost::fromString("1s").isEnoughFor(GCost::fromString("1d")));
    TS_ASSERT(!GCost::fromString("1$").isEnoughFor(GCost::fromString("1d")));

    TS_ASSERT(!GCost::fromString("1t").isEnoughFor(GCost::fromString("1m")));
    TS_ASSERT(!GCost::fromString("1d").isEnoughFor(GCost::fromString("1m")));
    TS_ASSERT( GCost::fromString("1m").isEnoughFor(GCost::fromString("1m")));
    TS_ASSERT(!GCost::fromString("1s").isEnoughFor(GCost::fromString("1m")));
    TS_ASSERT(!GCost::fromString("1$").isEnoughFor(GCost::fromString("1m")));

    TS_ASSERT(!GCost::fromString("1t").isEnoughFor(GCost::fromString("1s")));
    TS_ASSERT(!GCost::fromString("1d").isEnoughFor(GCost::fromString("1s")));
    TS_ASSERT(!GCost::fromString("1m").isEnoughFor(GCost::fromString("1s")));
    TS_ASSERT( GCost::fromString("1s").isEnoughFor(GCost::fromString("1s")));
    TS_ASSERT(!GCost::fromString("1$").isEnoughFor(GCost::fromString("1s")));

    TS_ASSERT(!GCost::fromString("1t").isEnoughFor(GCost::fromString("1$")));
    TS_ASSERT(!GCost::fromString("1d").isEnoughFor(GCost::fromString("1$")));
    TS_ASSERT(!GCost::fromString("1m").isEnoughFor(GCost::fromString("1$")));
    TS_ASSERT( GCost::fromString("1s").isEnoughFor(GCost::fromString("1$")));
    TS_ASSERT( GCost::fromString("1$").isEnoughFor(GCost::fromString("1$")));

    // Combinations including supply sale:
    TS_ASSERT( GCost::fromString("5t 3d 7m 99$").isEnoughFor(GCost::fromString("3tdm 42$")));
    TS_ASSERT( GCost::fromString("5t 3d 7m 99s").isEnoughFor(GCost::fromString("3tdm 42$")));
    TS_ASSERT(!GCost::fromString("5t 3d 7m 99s").isEnoughFor(GCost::fromString("4tdm 42$")));
    TS_ASSERT( GCost::fromString("5t 3d 7m 22s 22$").isEnoughFor(GCost::fromString("3tdm 42$")));
    TS_ASSERT(!GCost::fromString("5t 3d 7m 22s 22$").isEnoughFor(GCost::fromString("3tdm 52$")));
}
