/**
  *  \file ui/richparser.cc
  *  \brief Rich Text Parsing
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/richparser.h"
#include "util/unicodechars.h"
#include "cpluslib/misc.h"
#include "io/memstream.h"

/** Parse a color name. */
static UIColor::Color
parseColorName(const string_t& name)
{
    static const struct {
        const char* name;
        UIColor::Color color;
    } colors[] = {
        { "static", UIColor::tc_Static },
        { "green",  UIColor::tc_Green },
        { "yellow", UIColor::tc_Yellow },
        { "red",    UIColor::tc_Red },
        { "white",  UIColor::tc_White },
        { "blue",   UIColor::tc_Blue },
        { "dim",    UIColor::tc_Faded },
    };
    for (std::size_t i = 0; i < countof(colors); ++i) {
        if (name == colors[i].name) {
            return colors[i].color;
        }
    }
    return UIColor::tc_Static;
}

/**************************** RichParser ***************************/

/** Check for opening tag. If that tag is found, skips it from the token stream.
    \param what Tag to check for (case sensitive!) */
bool
RichParser::isOpen(const char* what)
{
    if (current == rdr.TagStart && rdr.getTag() == what) {
        next();
        return true;
    } else {
        return false;
    }
}

/** Parse text sequence. Parses a list of text elements until it encounters a closing tag.
    \param keepFormat true to keep format (inside <pre>), false to rewrap (outside <pre>)
    \return parsed text */
RichText
RichParser::parseText(bool keepFormat)
{
    /* Text consists of a, b, em, tt, key, font, where a and font have parameters. */
    bool haveSpace = true;
    RichText result;
    while (1) {
        if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
            /* EOF or error stop parsing, or surrounding tag ends */
            break;
        } else if (current == rdr.TagAttribute) {
            /* Spurious attribute from surrounding tag */
            next();
        } else if (current == rdr.Text) {
            /* Text */
            if (keepFormat) {
                result.append(rdr.getValue());
            } else {
                appendText(result, haveSpace, rdr.getValue());
            }
            next();
        } else {
            result.append(parseTextItem(keepFormat));
            next();
            haveSpace = false;
        }
    }
    return result;
}

/** Parse a text element.
    Parses a single tag in flow-text markup.
    \param keepFormat true to keep format (inside <pre>), false to rewrap (outside <pre>)
    \return parsed text */
RichText
RichParser::parseTextItem(bool keepFormat)
{
    if (isOpen("a")) {
        string_t target;
        while (current == rdr.TagAttribute) {
            if (rdr.getName() == "href") {
                target = rdr.getValue();
            }
            next();
        }
        return parseText(keepFormat).withAttribute(new RichTextLinkAttribute(target));
    } else if (isOpen("b")) {
        return parseText(keepFormat).withAttribute(RichTextStyleAttribute::Bold);
    } else if (isOpen("em")) {
        // Should be italic, but we don't have an italic font yet.
        // return parseText(keepFormat).withAttribute(RichTextStyleAttribute::Italic);
        return parseText(keepFormat).withAttribute(RichTextStyleAttribute::Bold);
    } else if (isOpen("u")) {
        return parseText(keepFormat).withAttribute(RichTextStyleAttribute::Underline);
    } else if (isOpen("tt")) {
        return parseText(keepFormat).withAttribute(RichTextStyleAttribute::Fixed);
    } else if (isOpen("kbd") || isOpen("key")) {
        return renderKeys(parseText(keepFormat).getText());
    } else if (isOpen("big")) {
        return parseText(keepFormat).withAttribute(RichTextStyleAttribute::Big);
    } else if (isOpen("small")) {
        return parseText(keepFormat).withAttribute(RichTextStyleAttribute::Small);
    } else if (isOpen("font")) {
        UIColor::Color c = UIColor::tc_Static;
        while (current == rdr.TagAttribute) {
            if (rdr.getName() == "color") {
                c = parseColorName(rdr.getValue());
            }
            next();
        }
        return parseText(keepFormat).withAttribute(new RichTextColorAttribute(c));
    } else if (current == rdr.TagStart) {
        /* An opening tag we don't understand; parse its content */
        next();
        return parseText(keepFormat);
    } else {
        /* Completely unknown, skip it */
        next();
        return RichText();
    }
}

/** Parse a piece of text.
    Simple all-in-one entry point. */
RichText
RichParser::parse()
{
    // FIXME: like the <p> handling, this is almost a copy of parseText.
    bool haveSpace = true;
    RichText result;
    while (1) {
        if (current == rdr.Eof || current == rdr.Error || current == rdr.TagEnd) {
            /* EOF or error stop parsing, or surrounding tag ends */
            break;
        } else if (current == rdr.TagAttribute) {
            /* Spurious attribute from surrounding tag */
            next();
        } else if (current == rdr.Text) {
            /* Text */
            appendText(result, haveSpace, rdr.getValue());
            next();
        } else if (isOpen("br")) {
            /* Newline */
            result.append("\n\n");
            while (current != rdr.Eof && current != rdr.Error && current != rdr.TagEnd) {
                next();
            }
            next();
            haveSpace = true;
        } else {
            result.append(parseTextItem(false));
            next();
            haveSpace = false;
        }
    }
    return result;
}

/** Skip a tag.
    Called with current being the unknown opening tag.
    Eats up balanced tags until that tag is closed. */
void
RichParser::parseUnknown()
{
    int nesting = 0;
    do {
        if (current == rdr.TagStart) {
            ++nesting;
        }
        if (current == rdr.TagEnd) {
            --nesting;
        }
        next();
    } while (current != rdr.Eof && nesting != 0);
}

/** Append string to RichText. This collapses multiple spaces into one, and strips newlines.
    \param out       [in/out] RichText to append to
    \param haveSpace [in/out] Space status tracking
    \param in        [in] Text to append */
void
RichParser::appendText(RichText& out, bool& haveSpace, const string_t& in)
{
    static const char SPACE[] = " \r\n\t";
    string_t::size_type a = 0;
    while (a < in.size()) {
        if (haveSpace) {
            /* out ends with a space. Skip all space in input. */
            a = in.find_first_not_of(SPACE, a);
            haveSpace = false;
        } else {
            /* out does not end with a space, so look for spaces in /in/. */
            string_t::size_type e = in.find_first_of(SPACE, a);
            if (e == in.npos) {
                out.append(string_t(in, a));
            } else {
                out.append(string_t(in, a, e-a));
                out.append(" ");
                haveSpace = true;
            }
            a = e;
        }
    }
}

/** Render a key specification. Documents contain \<kbd> tags for whole key strings, such
    as "Alt+A". This splits the key string into individual keys. */
RichText
RichParser::renderKeys(const string_t& name)
{
    RichText result;
    string_t::size_type p = 0;
    while (p < name.size()) {
        /* Delimiters:
           "-" and "+" for key combinations "Alt-K"
           "/" for alternatives "Up/Down"
           ",.;: " for punctuation "Up, Down" */
        string_t::size_type e = name.find_first_of("-+/,.;: ", p+1);
        if (e == string_t::npos) {
            /* String ends */
            result.append(RichText(string_t(name, p)).withAttribute(RichTextStyleAttribute::Key));
            break;
        }

        /* String does not end after key, but may have punctuation */
        result.append(RichText(string_t(name, p, e-p)).withAttribute(RichTextStyleAttribute::Key));
        p = e;
        if (name[p] == '.') {
            e = name.find_first_not_of(". ", p+1);
        } else {
            e = name.find_first_not_of(" ", p+1);
        }
        if (e == string_t::npos) {
            /* String ends with punctuation */
            result.append(string_t(name, p));
            break;
        }

        /* String does not end after punctuation. As a special case, if it's a
           normal dash, turn that into a (shorter) hyphen. */
        if (e == p+1 && name[p] == '-') {
            result.append(UTF_HYPHEN);
        } else {
            result.append(string_t(name, p, e-p));
        }
        p = e;
    }

    return result;
}

RichText
RichParser::textFromXml(string_t source)
{
    MemoryStream ms(source);
    XmlReader rdr(ms);
    RichParser p(rdr);
    p.next();
    return p.parseText(true);
}
