/**
  *  \file ui/widgets/tree.cc
  *  \brief Tree widget
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "ui/widgets/tree.h"

/** Constructor.
    \param id Widget Id
    \param lines Number of lines to display
    \param width Width in pixels */
UITreeWidget::UITreeWidget(int id, int lines, int width)
    : UIListbox(id, lines, width, 0)
{ }

UITreeWidget::~UITreeWidget()
{ }

/** Get item index of an item's parent.
    \param itemIndex item whose parent to query (index into items) */
uint32_t
UITreeWidget::getParentOf(uint32_t itemIndex)
{
    int32_t level = items[itemIndex]->level;
    while (itemIndex > 0 && items[itemIndex]->level >= level) {
        --itemIndex;
    }
    return itemIndex;
}

/** Get line of an item. When the item is not visible, returns the line
    of the closest preceding visible item.
    \param itemIndex item whose line to query (index into items) */
uint32_t
UITreeWidget::getLineOf(uint32_t itemIndex)
{
    /* We could do binary search, but this way it's simpler */
    uint32_t i = 0;
    while (i+1 < lineToItem.size() && lineToItem[i+1] <= itemIndex) {
        ++i;
    }
    return i;
}

/** Check whether an item has children (=is a node, not a leaf). */
bool
UITreeWidget::hasChildren(uint32_t itemIndex) const
{
    return itemIndex+1 < items.size()
        && items[itemIndex+1]->level > items[itemIndex]->level;
}

/** Add item. \overload */
void
UITreeWidget::addItem(int32_t id, int32_t level, string_t label)
{
    addItem(id, level, label, false);
}

/** Add item.
    \param id    Identificator for this item.
    \param level Nesting level. The root level is 0. Children have their parent's level plus one.
                 An item is followed by its children which have higher level than itself,
                 and it is preceded (directly or indirectly) by its parent, grandparent, etc.
                 with lower levels.
    \param label Item label (user-visible text)
    \param open  For nodes, whether it is initially open or not (default: false). For leaves, ignored. */
void
UITreeWidget::addItem(int32_t id, int32_t level, string_t label, bool open)
{
    std::auto_ptr<Item> p(new Item);
    p->id = id;
    p->level = level;
    p->open = open;
    p->label = label;
    items.push_back(p.release());
}

/** Update display after something in the tree model changed. Call
    this after using addItem(). We call it ourselves after opening or
    closing items. */
void
UITreeWidget::updateAfterModification()
{
    /* Remember which item is active */
    uint32_t activeItem = (getCurrent() >= 0 && getCurrent() < int32_t(lineToItem.size())) ? lineToItem[getCurrent()] : 0;

    /* Rebuild lineToItem */
    lineToItem.clear();
    uint32_t i = 0;
    while (i < items.size()) {
        int32_t level = items[i]->level;
        lineToItem.push_back(i);
        if (items[i]->open) {
            /* It's open, so copy the children as well */
            ++i;
        } else {
            /* It's closed, so skip all children */
            ++i;
            while (i < items.size() && items[i]->level > level) {
                ++i;
            }
        }
    }

    /* Now, find out new place to put cursor to. This implicitly handles the
       pathological case of an empty lineToItem, because getLineOf returns
       0 in that case. */
    uint32_t line = getLineOf(activeItem);
    while (line != 0 && lineToItem[line] != activeItem) {
        activeItem = getParentOf(activeItem);
        line = getLineOf(activeItem);
    }

    /* Configure list box */
    setSize(lineToItem.size());
    scrollTo(line, 1);
    drawWidget();
    sig_moved.raise(getId());
}

void
UITreeWidget::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    uint32_t index = getTop() + from;
    for (int i = from; i < to; ++i) {
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        if (index < lineToItem.size()) {
            /* Find indentation */
            const uint32_t slot = lineToItem[index];
            const Item& it = *items[slot];
            const int wi = font_ems[FONT_NORMAL];
            const int indent = wi * it.level + 5;

            area.consumeX(indent);
            ctx.setTextAlign(0, 1);

            /* +/- */
            if (hasChildren(slot)) {
                ctx.setColor(UIColor::tc_Faded);
                outTextF(ctx, area.splitX(wi), it.open ? "-" : "+");
            } else {
                area.consumeX(wi);
            }

            /* Text */
            ctx.setColor(UIColor::tc_Static);
            drawItem(ctx, area, it);
            ++index;
        }
    }
}

/** Draw an item. This can be overridden to provide a different item format. */
void
UITreeWidget::drawItem(GfxContext& ctx, GfxRect area, const Item& it)
{
    outTextF(ctx, area, it.label);
}

bool
UITreeWidget::handleEvent(const UIEvent& event, bool second_pass)
{
    /* Handle key events when the widget state allows and the cursor is in a valid position */
    if (hasState(st_Focused) && !hasState(st_Disabled)
        && getCurrent() >= 0 && getCurrent() < int32_t(lineToItem.size())
        && event.type == ev_Key)
    {
        const uint32_t slot = lineToItem[getCurrent()];
        Item& item = *items[slot];
        switch (event.key.code) {
         case SDLK_LEFT:
            /* close if it is open, otherwise go to parent */
            if (item.open && hasChildren(slot)) {
                item.open = false;
                updateAfterModification();
            } else if (slot > 0) {
                scrollTo(getLineOf(getParentOf(slot)), 1);
            }
            return true;
         case SDLK_RIGHT:
            /* open if it is closed, otherwise go to child */
            if (hasChildren(slot)) {
                if (!item.open) {
                    item.open = true;
                    updateAfterModification();
                } else {
                    scrollTo(getCurrent()+1, 1);
                }
            }
            return true;
         case ' ':
            /* open/close item, do not scroll */
            if (hasChildren(slot)) {
                item.open = !item.open;
                updateAfterModification();
            }
            return true;
         case '*':
            /* toggle current, and set all its children AND LATER SIBLINGS to same state */
            if (hasChildren(slot)) {
                item.open = !item.open;
                for (uint32_t i = slot+1; i < items.size() && items[i]->level >= item.level; ++i)
                    items[i]->open = item.open;
                updateAfterModification();
            }
            return true;
        }
    }
    return UIListbox::handleEvent(event, second_pass);
}

void
UITreeWidget::onItemClick(int item, int x)
{
    if (item >= 0 && item < int(lineToItem.size())) {
        /* Check for click on a +/- */
        const uint32_t slot = lineToItem[item];
        Item& item = *items[slot];
        const int wi = font_ems[FONT_NORMAL];
        const int indent = wi * item.level;
        if (x >= indent && x < indent + wi + 5 && hasChildren(slot)) {
            item.open = !item.open;
            updateAfterModification();
        }
    }
}

/** Get Id of tree item. */
int32_t
UITreeWidget::getItemId(uint32_t itemIndex) const
{
    if (itemIndex < items.size()) {
        return items[itemIndex]->id;
    } else {
        return 0;
    }
}

/** Get item index of currently-selected tree item. */
uint32_t
UITreeWidget::getCurrentItemIndex() const
{
    return getItemIndex(getCurrent());
}

/** Get item index of a tree item by line. */
uint32_t
UITreeWidget::getItemIndex(int line) const
{
    if (line >= 0 && line < int(lineToItem.size())) {
        return lineToItem[line];
    } else {
        return 0;
    }
}

/** Toggle state of specified tree item. */
void
UITreeWidget::toggleItem(uint32_t itemIndex)
{
    if (itemIndex < items.size() && hasChildren(itemIndex)) {
        items[itemIndex]->open = !items[itemIndex]->open;
        updateAfterModification();
    }
}
