/**
  *  \file util/conf.cc
  *  \brief Configuration Data Management
  *
  *  \todo Exception handling must be correctly specified and implemented.
  *  Right now we throw range_error's or ignore errors.
  *
  *  \todo Associate some "source" parameter instead of just a
  *  "was_set" flag with each option, to implement things like "this
  *  option was set in the system configuration file" or "save only
  *  the options modified on the config screen to the game config
  *  file".
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "util/conf.h"
#include "util/except.h"
#include "util/nls.h"
#include "util/misc.h"
#include "cpluslib/format.h"

/* Remove comment from an option string. */
void
removeComment(string_t& str)
{
    string_t::size_type n = str.find('#');
    if (n != str.npos) {
        str.erase(n);
    }
}

/********************************* Config ********************************/

/** Constructor. Make blank object. */
Config::Config()
{ }

/** Destructor. */
Config::~Config()
{ }

/** Add option.
    \param option The option. Its lifetime must be the same or longer as the
                  Config object's. */
void
Config::addConfigOption(ConfigOption& option)
{
    options.push_back(&option);
}

/** Get option by name.
    \param name Name of config option
    \return the option, or null if none matches */
ConfigOption*
Config::getConfigOption(const string_t name)
{
    for (std::vector<ConfigOption*>::iterator i = options.begin(); i != options.end(); ++i)
        if (strCaseCmp(name, (*i)->getName()) == 0)
            return *i;
    return 0;
}

/** Mark all options as unset. */
void
Config::markAllOptionsUnset()
{
    for (std::vector<ConfigOption*>::iterator i = options.begin(); i != options.end(); ++i)
        (*i)->setSource(ConfigOption::srcDefault);
}

/****************************** ConfigOption *****************************/

/** Constructor. Constructs an object and adds it to the container.
    \param container Config object
    \param name Name of option. Lifetime must be the same or longer as
                this ConfigOption object's, i.e. static. */
ConfigOption::ConfigOption(Config& container, const char* name)
    : name(name), source(srcDefault)
{
    container.addConfigOption(*this);
}

/** Mark this option updated. If this option is set at a level below \c src,
    upgrade it to that level. In particular, this moves a default option to
    the respective config file. */
void
ConfigOption::markUpdated(Source src)
{
    if (this->source < src) {
        this->source = src;
    }
}

/** Set this option, and mark it updated if it changed.
    This is the preferred way to set options from user input. */
void
ConfigOption::setAndMarkUpdated(string_t value, Source src)
{
    if (value != toString()) {
        set(value);
        markUpdated(src);
    }
}

/****************************** ValueParser ******************************/

/** Parse comma-separated list into array.
    \param value [in] String value
    \param array [out] The array
    \param dim   [in] Number of elements in array */
void
ValueParser::parseArray(string_t value, int32 array[], int dim)
{
    int count = 0;
    int32 last_value;
    string_t::size_type n;
    while (count < dim-1 && (n = value.find(',')) != string_t::npos) {
        last_value = parse(strTrim(value.substr(0, n)));
        value.erase(0, n+1);
        array[count++] = last_value;
    }
    last_value = parse(strTrim(value));
    array[count++] = last_value;

    while (count < dim)
        array[count++] = last_value;
}

string_t
ValueParser::toStringArray(const int32 array[], int dim)
{
    string_t result = toString(array[0]);
    for (int i = 1; i < dim; ++i) {
        result += ",";
        result += toString(array[i]);
    }
    return result;
}

/***************************** ValueIntParser ****************************/

ValueIntParser ValueIntParser::instance;

/** Constructor. */
ValueIntParser::ValueIntParser()
{ }

/** Parse integer. */
int32
ValueIntParser::parse(string_t value)
{
    int32_t result;
    string_t::size_type pos;
    if (!stringValue(value, result, pos) && !stringValue(value.substr(0, pos), result)) {
        throw std::range_error(_("Invalid number"));
    }
    return result;
}

string_t
ValueIntParser::toString(int32 value)
{
    return format("%d", value);
}

/**************************** ValueBoolParser ****************************/

ValueBoolParser ValueBoolParser::instance;

ValueBoolParser::ValueBoolParser()
{ }

int32
ValueBoolParser::parse(string_t value)
{
    if (stringMatch("No", value) || stringMatch("False", value))
        return 0;
    if (stringMatch("Yes", value) || stringMatch("True", value))
        return 1;
    if (stringMatch("Allies", value))
        return 2;
    if (stringMatch("External", value))
        return 3;

    int32 tmp;
    if (stringValue(value, tmp))
        return tmp;

    return 1; /* arbitrary */
}

string_t
ValueBoolParser::toString(int32 value)
{
    switch (value) {
     case 0:
        return "No";
     case 1:
     default: /* arbitrary */
        return "Yes";
     case 2:
        return "Allies";
     case 3:
        return "External";
    }
}

/*************************** ValueBitsetParser ***************************/

static bool
nextWord(const char*& tpl, string_t& word)
{
    using namespace std;
    if (*tpl == 0) {
        // end reached
        return false;
    } else if (const char* p = strchr(tpl, ',')) {
        // intermediate word
        word.assign(tpl, p - tpl);
        tpl = p+1;
        return true;
    } else {
        // final word
        word.assign(tpl);
        tpl += strlen(tpl);
        return true;
    }
}

/** Constructor.
    \param tpl Template naming all the bits, in the form "zero,one,two,four,eight". */
ValueBitsetParser::ValueBitsetParser(const char* tpl)
    : tpl(tpl)
{ }

int32
ValueBitsetParser::parse(string_t value)
{
    int32 result = 0;
    do {
        string_t item = strTrim(strFirst(value, ","));
        int32 tmp;
        if (item.size() == 0) {
            // blank field
        } else if (stringValue(item, tmp)) {
            // numeric field
            result |= tmp;
        } else {
            // word field
            const char* p = tpl;
            string_t ele;
            int32 bit = 1;
            while (nextWord(p, ele)) {
                if (strCaseCmp(ele, item) == 0) {
                    result |= bit;
                    break;
                }
                bit <<= 1;
            }
        }
    } while (strRemove(value, ","));
    return result;
}

string_t
ValueBitsetParser::toString(int32 value)
{
    string_t result;
    const char* p = tpl;
    string_t ele;
    int32 bit = 1;
    while (nextWord(p, ele)) {
        if (value & bit) {
            if (result.size())
                result.append(",");
            result.append(ele);
        }
        bit <<= 1;
    }
    return result;
}

/**************************** ValueEnumParser ****************************/

ValueEnumParser::ValueEnumParser(const char* tpl)
    : tpl(tpl)
{ }

int32_t
ValueEnumParser::parse(string_t value)
{
    /* Try to parse as enum element */
    const char* p = tpl;
    string_t ele;
    int32_t counter = 0;
    while (nextWord(p, ele)) {
        if (strCaseCmp(value, ele) == 0) {
            return counter;
        }
        ++counter;
    }
    throw std::range_error(_("Invalid number"));
}

string_t
ValueEnumParser::toString(const int32_t value)
{
    const char* p = tpl;
    string_t ele;
    int32_t counter = 0;
    while (nextWord(p, ele)) {
        if (counter == value) {
            return ele;
        }
        ++counter;
    }

    /* Fallback, just in case */
    return format("%d", value);
}

/************************ ConfigIntArrayBaseOption ***********************/

bool
ConfigIntArrayBaseOption::isAllTheSame() const
{
    const int32_t* array = getArray();
    int limit = getLimit();
    if (limit != 0) {
        int value = array[0];
        for (int i = 1; i < limit; ++i) {
            if (array[i] != value) {
                return false;
            }
        }
    }
    return true;
}

/*************************** ConfigStringOption **************************/

ConfigStringOption::ConfigStringOption(Config& container, const char* name)
    : ConfigOption(container, name)
{ }

ConfigStringOption::~ConfigStringOption()
{ }

void
ConfigStringOption::set(string_t value)
{
    this->value = value;
}

string_t
ConfigStringOption::toString() const
{
    return value;
}

string_t
ConfigStringOption::operator()() const
{
    return value;
}

/************************ ConfigStringArrayOption ************************/

ConfigStringArrayOption::ConfigStringArrayOption(Config& container, const char* name, int first_index, int num_elements)
    : ConfigOption(container, name),
      first_index(first_index),
      num_elements(num_elements),
      data(num_elements)
{ }

ConfigStringArrayOption::~ConfigStringArrayOption()
{ }

void
ConfigStringArrayOption::set(string_t value)
{
    size_t pos = 0;
    for (size_t i = 0; i < data.size(); ++i) {
        size_t next = value.find(',', pos);
        if (next == string_t::npos) {
            data[i] = strTrim(value.substr(pos));
        } else {
            data[i] = strTrim(value.substr(pos, next - pos));
            pos = next+1;
        }
    }
}

string_t
ConfigStringArrayOption::toString() const
{
    string_t result = data[0];
    for (size_t i = 1; i < data.size(); ++i) {
        result += ",";
        result += data[i];
    }
    return result;
}

string_t
ConfigStringArrayOption::operator()(int index) const
{
    index -= first_index;
    if (index < 0 || index >= num_elements) {
        return string_t();
    } else {
        return data[index];
    }
}

/****************************** ConfigAlias ******************************/


ConfigAlias::ConfigAlias(Config& container, const char* name, ConfigOption& other)
    : ConfigOption(container, name),
      other(other)
{ }

void
ConfigAlias::set(string_t value)
{
    other.set(value);
}

string_t
ConfigAlias::toString() const
{
    return other.toString();
}
